/*
  +-------------------------------------+
  |                                     |
  |   Pseudo RSH daemon for Winboard.   |
  |   Written by D.A.Newman, Dec-99.    |
  |                                     |
  +-------------------------------------+
*/
#include <stdlib.h>
#include <process.h>
#include <stdio.h>
#include <direct.h>
#include <winsock.h>
#include "timer.h"
#include "print.h"
#include "socks.h"
#include "serror.h"

int verbose = 0;

const int CREATE_PIPE_FAILURE = 101;
const int DUP_PIPE_FAILURE = 102;
const int CREATE_PROCESS_FAILURE = 103;
const int LAUNCH_THREAD_FAILURE = 104;

class engine_t {
  private:
    HANDLE process_handle;
    HANDLE command_in;
    HANDLE command_out;
    HANDLE stdin_thread;
    HANDLE stdout_thread;
    socket_t stdio_socket;
    int running;
    void launch( char *command);
    // I/O thread proceedures:
    static unsigned __stdcall input_proc( void *tparam);
    static unsigned __stdcall output_proc( void *tparam);
  public:
    engine_t( char *cmdline, socket_t &socket);
    ~engine_t();
    void wait_till_done();
};

engine_t::engine_t( char *cmdline, socket_t &socket)
  : process_handle( 0), command_in( 0), command_out( 0),
    stdin_thread( 0), stdout_thread( 0), stdio_socket( socket), running( 0)
{
    //
    // Split the directory portion of the command line
    // off so we can cd to the appropriate directory.
    //
    static char drive[_MAX_DRIVE];
    static char dir[_MAX_DIR];
    static char fname[_MAX_FNAME];
    static char ext[_MAX_EXT];
    static char path[_MAX_PATH];
    static char prog[_MAX_PATH];
    _splitpath( cmdline, drive, dir, fname, ext);
    _makepath( path, drive, dir, NULL, NULL);
    _makepath( prog, NULL, NULL, fname, ext);

    if( verbose ) tprintf("Path: %s\n", path);
    if( verbose ) tprintf("Prog: %s\n", prog);

    static char current_dir[_MAX_PATH];
    getcwd( current_dir, _MAX_PATH);
    if( verbose ) tprintf("Current dir: %s\n", current_dir);

    chdir( path);

    getcwd( current_dir, _MAX_PATH);
    if( verbose ) tprintf("New dir: %s\n", current_dir);

    //
    // Launch engine.
    //
    launch( prog);

    //
    // Start thread to handle engine input.
    //
    unsigned worker_id1;
    stdin_thread = (HANDLE)_beginthreadex( NULL, 0,
      input_proc, this, 0, &worker_id1);

    if( stdin_thread == 0 ) {
        tprintf("Error: failure launching engine input thread\n");
        throw LAUNCH_THREAD_FAILURE;
    }

    if( verbose ) tprintf("Launched stdin thread....\n");

    //
    // Start thread to handle engine output.
    //
    unsigned worker_id2;
    stdout_thread = (HANDLE)_beginthreadex( NULL, 0,
      output_proc, this, 0, &worker_id2);

    if( stdout_thread == 0 ) {
        tprintf("Error: failure launching engine output thread\n");
        throw LAUNCH_THREAD_FAILURE;
    }

    if( verbose ) tprintf("Launched stdout thread....\n");
}

engine_t::~engine_t()
{
    WaitForSingleObject( stdin_thread, INFINITE);
    WaitForSingleObject( stdout_thread, INFINITE);
    if( stdin_thread != 0 ) CloseHandle( stdin_thread);
    if( stdout_thread != 0 ) CloseHandle( stdout_thread);
    if( command_in != 0 ) CloseHandle( command_in);
    if( command_out != 0 ) CloseHandle( command_out);
    if( process_handle != 0 ) CloseHandle( process_handle);
}

//
// Launch the Chess engine: create the pipes to connect engine's
// stdio to this process, and spawn off the engine.  (Much modified
// from a chunk of code from Winboard...)
//
void engine_t::launch( char *command)
{
    //
    // We fill this out so that the child process will inherit
    // the inheritable handles.
    //
    SECURITY_ATTRIBUTES sa;
    sa.nLength = sizeof( sa);
    sa.bInheritHandle = TRUE;
    sa.lpSecurityDescriptor = NULL;

    //
    // The steps for redirecting child's STDOUT:
    //    1. Create anonymous pipe to be STDOUT for child.
    //    2. Create a noninheritable duplicate of read handle,
    //        and close the inheritable read handle.
    //

    //
    // Create pipe for command's standard out.
    //
    HANDLE stdout_read, stdout_write;
    BOOL ok = CreatePipe( &stdout_read, &stdout_write, &sa, 0);
    if( !ok ) {
        tprintf("Failed to create command stdout pipe\n");
        //tprintf("System: %s\n", system_error_string( GetLastError()));
        throw CREATE_PIPE_FAILURE;
    }

    //
    // Duplicate read handle of stdout pipe with inherit
    // flag set to FALSE so that it can't be inherited
    // by the command process.
    //
    HANDLE curproc = GetCurrentProcess();
    HANDLE temp;
    ok = DuplicateHandle( curproc, stdout_read,
                          curproc, &temp,
                          0, FALSE, DUPLICATE_SAME_ACCESS);
    command_out = temp;

    if( !ok ) {
        tprintf("Failed to duplicate pipe read handle\n");
        //       "System: %s", system_error_string( GetLastError()));
        throw DUP_PIPE_FAILURE;
    }

    CloseHandle( stdout_read);

    //
    // The steps for redirecting child's STDIN:
    //     1. Create anonymous pipe to be STDIN for child.
    //     2. Create a noninheritable duplicate of write handle,
    //         and close the inheritable write handle.
    //

    //
    // Create a pipe for the command's stdin.
    //
    HANDLE stdin_read, stdin_write;
    ok = CreatePipe( &stdin_read, &stdin_write, &sa, 0);
    if( !ok ) {
        tprintf("Failed to create command stdin pipe\n");
        //       "System: %s", system_error_string( GetLastError()));
        throw CREATE_PIPE_FAILURE;
    }

    //
    // Duplicate the write handle to the stdin pipe in such
    // a way that it cannot be inherited.  That way this process
    // will have it but the command won't.
    //
    ok = DuplicateHandle( curproc, stdin_write,
                          curproc, &temp,
                          0, FALSE, DUPLICATE_SAME_ACCESS);
    command_in = temp;
    if( !ok ) {
        tprintf("Failed to duplicate pipe write handle\n");
        //       "System: %s", system_error_string( GetLastError()));
        throw DUP_PIPE_FAILURE;
    }

    CloseHandle( stdin_write);

    //
    // Now we launch the command.
    //
    STARTUPINFO startup_info;
    ZeroMemory( &startup_info, sizeof( startup_info));
    startup_info.cb = sizeof(STARTUPINFO);
    startup_info.dwFlags = STARTF_USESTDHANDLES;
    startup_info.hStdInput = stdin_read;
    startup_info.hStdOutput = stdout_write;
    startup_info.hStdError = stdout_write;

    PROCESS_INFORMATION process_info;
    ok = CreateProcess( NULL,              // application name
                        command,           // command line
                        NULL,              // process security attributes
                        NULL,              // primary thread security attrs
                        TRUE,              // handles are inherited
                        DETACHED_PROCESS,  // | CREATE_NEW_PROCESS_GROUP,
                        NULL,              // use parent's environment
                        NULL,
                        &startup_info,
                        &process_info); // gets the process information.

    DWORD err = GetLastError();
    if( err != 0 ) printf("Error mes: %s\n", system_error_string( err));

    //
    // Now that the command has (probably) started we can close the
    // the handles for the ends of the pipes that the command
    // is going to use.
    //
    CloseHandle( stdin_read);
    CloseHandle( stdout_write);

    if( !ok ) {
        tprintf("Failed to spawn command (%s) process\n", command);
        //tprintf("System: %s\n", system_error_string( err));
        throw CREATE_PROCESS_FAILURE;
    }

    process_handle = process_info.hProcess;

    running = 1;
}

//
// Wait for engine to quit.
//
void engine_t::wait_till_done()
{
    if( running ) WaitForSingleObject( process_handle, INFINITE);
    running = 0;
}

//
// This routine takes parses the data gotten from Winboard via the
// socket connection and returns one datum/command at a time.
// This solves the problem I had with the commercial rshell daemon
// I tried.  (It ate the first few Winboard commands after it got
// the command line because they were all stuffed into the same
// packet...)
//
int get_line( socket_t &s, char *buf, int /*len*/)
{
    static char buffer[4096];
    static int n = 0;
    if( buf == 0 ) { n = 0; return 0; } // For resetting...

    while( 1 ) {
        int len;
        int gotone = 0;
        for( len = 0; len < n; len++ ) {
            if( buffer[len] == '\0'
              || buffer[len] == '\n' ) {
                len++;
                gotone = 1;
                break;
            }
        }
        if( gotone ) {
            for( int j = 0; j < len; j++ ) buf[j] = buffer[j];
            for( int k = 0; j < n; j++, k++ ) buffer[k] = buffer[j];
            n = k;
            buf[len] = '\0';
            return len;
        }
        n = n + s.recv( buffer+n, 4096);
        //tprintf("nnnnn: %d\n", n);
    }
}

//
// Thread proceedure for handing stdin to the engine.  It receives
// Winboard commands from Winboard via a socket connection and writes
// them to the engine through its stdin pipe.
//
unsigned __stdcall engine_t::input_proc( void *tparam)
{
    engine_t *engine = (engine_t *)tparam;
    const DWORD BFSZ = 255;
    static char buf[ BFSZ+1];

    try {
        //
        // This loop recvs until a line terminator makes it into the
        // buffer.  Then it sends a message to hwnd and waits till
        // it gets a signal to continue.
        //
        DWORD nchars = 0;
        while( engine->running ) {
            int nread = get_line( engine->stdio_socket, buf, BFSZ);
            #if 0
            tprintf("Winboard command: %s", buf);
            #endif
            if( !engine->running ) break;
            unsigned long nwritten;
            WriteFile( engine->command_in, buf, strlen( buf), &nwritten, NULL);
        }
        return 0;
    } catch( int &except ) {
        if( except == RECV_FAILURE && engine->running ) {
            tprintf("Error: connection to opponent interface broken\n");
            if( verbose ) tprintf("Sending a \"quit\" to the engine\n");
            unsigned long nw;
            WriteFile( engine->command_in, "quit\n", 5, &nw, NULL);
        }
        return -1;
    }
}

//
// Thread proceedure for handing stdout from the engine.  It reads
// Winboard commands from the engine's stdout pipe and sends them
// on to Winboard via a socket connection.
//
unsigned __stdcall engine_t::output_proc( void *tparam)
{
    engine_t *engine = (engine_t *)tparam;
    const DWORD BFSZ = 255;
    static char buf[ BFSZ+1];

    try {
        //
        // This loop reads until a line terminator makes it into the
        // buffer.  Then it sends the full command to Winboard over
        // the socket connection.
        //
        DWORD nread;
        DWORD nchars = 0;
        while( engine->running ) {

            BOOL ok = ReadFile( engine->command_out, buf+nchars,
              BFSZ-nchars, &nread, NULL);
            if( !engine->running ) break;
            if( !ok && engine->running ) {
                tprintf("Error: pipe to engine's stdout broken\n");
                tprintf("System: %s\n", system_error_string( GetLastError()));
                return -1;
            }
            nchars += nread;
            buf[nchars] = '\0';
            while( strchr( buf, '\r') || strchr( buf, '\n') ) {
                char *bp = strtok( buf, "\r\n");
                if( bp == NULL ) {
                    nchars = 0;
                    break;
                }
                DWORD len = strlen( bp);
                static char buffer[1024];
                strncpy( buffer, bp, 1024);
                buffer[len] = '\n';
                if( !engine->running ) break;
                int nsent = engine->stdio_socket.send( buffer, len+1);
                //printf("buffer(%d,%d): %s\n", len, nsent, bp);
                bp = strtok( NULL, "");
                if( bp == NULL ) {
                    nchars = 0;
                    break;
                }
                while( bp[0] == '\n' || bp[0] == '\r' ) bp++;
                nchars = strlen( bp);
                if( nchars ) memmove( buf, bp, nchars);
                buf[nchars] = '\0';
            }
        }
        return 0;
    } catch( int &except ) {
        if( except == SEND_FAILURE && engine->running ) {
            tprintf("Error: connection to opponent interface broken\n");
        }
        return -1;
    }
}

//
// Listen for rsh connection.  When connection is made launch off Chess
// engine and pass Winboard commands back and forth.  When done, return
// various resources, loop around, and start over.
//
int main()
{
    const int BFSZ = 1024;
    static char buf[ BFSZ];
    init_time();

    while( 1 ) {

        try {
            //
            // Initialize WinSock.
            //
            wsockets_t ws;

            debug_print = 1;
            print_localhost_info();
            debug_print = 0;

            //
            // Listen on port 514.
            //
            socket_t listen_socket;

            listen_socket.listen( 514, 1);

            tprintf("Ready...\n");

            socket_t stdio_socket = listen_socket.accept();

            //
            // No further connections allowed.
            //
            listen_socket.close();

            //
            // Reset get_line().
            //
            get_line( stdio_socket, 0, 0);

            //
            // Receive the port number for stderr output which
            // we don't fully connect up, but just do this to
            // satisfy the rsh stuff in Winboard.
            //
            int nchars = get_line( stdio_socket, buf, BFSZ);
            if( verbose ) tprintf("Got port: %s\n", buf);
            unsigned short port = (unsigned short)atoi( buf);
            socket_t stderr_socket;
            //
            // Get name of host that we are connecting to.
            //
            const int NMSZ = 64;
            static char name[ NMSZ];
            stdio_socket.getpeername( name, NMSZ);
            //
            // Connect up to stderr port.
            //
            stderr_socket.connect( name, port);

            //
            // Now we get the usernames which we just toss.
            //
            nchars = get_line( stdio_socket, buf, BFSZ);
            if( verbose ) tprintf("Got username: %s\n", buf);
            nchars = get_line( stdio_socket, buf, BFSZ);
            if( verbose ) tprintf("Got username2: %s\n", buf);

            //
            // Get the command line.
            //
            nchars = get_line( stdio_socket, buf, BFSZ);
            if( verbose ) tprintf("Got command: %s\n", buf);

            engine_t engine( buf, stdio_socket);

            tprintf("Engine launched.\n");

            //
            // Wait for engine to quit.
            //
            engine.wait_till_done();

            tprintf("Engine exited...\n");

        } catch( int &except ) {

            if( except == WINSOCK_FAILURE )
                tprintf("WinSock init failure\n");

            else if( except == BAD_SOCKET )
                tprintf("Bad socket\n");

            else if( except == BIND_FAILURE )
                tprintf("Bind failure\n");

            else if( except == LISTEN_FAILURE )
                tprintf("Listen failure\n");

            else if( except == HOSTNAME_LOOKUP_FAILURE )
                tprintf("Hostname lookup failure\n");

            else if( except == IPADDRESS_LOOKUP_FAILURE )
                tprintf("IP Address lookup failure\n");

            else if( except == CONNECT_FAILURE )
                tprintf("Connect failure\n");

            else if( except == ACCEPT_FAILURE )
                tprintf("Accept failure\n");

            else if( except == SEND_FAILURE )
                tprintf("Send failure\n");

            else if( except == RECV_FAILURE )
                tprintf("Failure receiving on socket\n");

            else
                tprintf("Unhandled exception\n");

            return -1;

        }

    }

}
